#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# Martin Still (SAAO)    1.0          16 Oct 2006

# mosaics SALT CCD amplifiers into a single image

from pyraf import iraf 

import saltprint, saltkey, saltio, saltstring
import os, time
import numpy, pyfits

# -----------------------------------------------------------
# core routine

def saltmosaic(images,outimages,outpref,geomfile,interp,cleanup,clobber,logfile,verbose,status):

    status = 0

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    history = 'SALTMOSAIC -- '
    history += 'images='+images+' '
    history += 'outimages='+outimages+' '
    history1 = 'outpref='+outpref+' '
    history1 += 'geomfile='+geomfile+' '
    history1 += 'interp='+interp+' '
    history1 += 'logfile='+logfile+' '
    yn = 'n'
    if (clobber): yn = 'y'
    history1 += 'clobber='+yn+' '
    yn = 'n'
    if (verbose): yn = 'y'
    history1 += 'verbose='+yn
    history += history1
    saltprint.log(logfile,history,verbose)

# start time

    saltprint.time('\nSALTMOSAIC -- started at',logfile,verbose)
    saltprint.log(logfile,' ',verbose)

# are the arguments defined
   
    if (status == 0): status = saltio.argdefined('images',images,logfile)
    if (status == 0): status = saltio.argdefined('geomfile',geomfile,logfile)
    if (status == 0): status = saltio.argdefined('logfile',logfile,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# load output name list: @list, * and comma separated
   
    outimages = outimages.strip()
    outpref = outpref.strip()
    if (status == 0 and len(outpref) == 0 and len(outimages) == 0):
        message = '\nERROR -- SALTMOSAIC: output file(s) not specified'
        status = saltprint.err(logfile,message)

# test output @filelist exists

    if (status == 0 and len(outimages) > 0 and outimages[0] == '@'): 
        status = saltio.listexists('Output',outimages,logfile)

# parse list of output files

    if (status == 0):
        outfiles, status = saltio.listparse('Output image',outimages,outpref,infiles,'',logfile)

# are input and output lists the same length?

    if (status == 0): status = saltio.comparelists(infiles,outfiles,'Input','output',logfile)

# do the output files already exist?

    if (status == 0 and not clobber): status = saltio.filesexist(outfiles,'','w',logfile)

# does CCD geometry definition file exist

    geomfilefile = geomfile.strip()
    if (status == 0): status = saltio.fileexists(geomfile,logfile)

# read geometry definition file

    gap = 0
    xshift = [0, 0]
    yshift = [0, 0]
    rotation = [0, 0]
    if (status == 0): gap, xshift, yshift, rotation, status = saltio.readccdgeom(geomfile,logfile,status)

# open each raw image file

    ifil = 0
    for infile in infiles:
	if (status == 0):
	    outfile = outfiles[ifil]
	    outpath = outfile.rstrip(os.path.basename(outfile))
	    if (len(outpath) == 0): outpath = '.'
	    struct, status = saltio.openfits(infile,logfile)

# identify instrument

	if (status == 0): instrume,keyprep,keygain,keybias,keyxtalk,keyslot,status = \
	    saltkey.instrumid(struct,infile,logfile)

# has file been prepared?

	if (status == 0): status = saltkey.prepare(struct,infile,keyprep,logfile)

# has file already been mosaiced?

	if (status == 0): 
	    status = 2
	    keymosaic = ''
	    if (instrume == 'RSS' or instrume == 'PFIS'): keymosaic = 'PMOSAIC'
	    if (instrume == 'SALTICAM'): keymosaic = 'SMOSAIC'
	    try:
		value = struct[0].header[keymosaic]
	    except:
		status = 0
	if (status == 2): 
	    message = 'ERROR -- SALTMOSAIC: ' + infile + ' has already been mosaiced\n'
	    status = saltprint.err(logfile,message)

# how many amplifiers?

	if (status == 0): 
	    nccds, status = saltkey.get('NCCDS',struct[0],infile,logfile)
	    amplifiers = nccds * 2

# CCD geometry coefficients

	if (status == 0 and (instrume == 'RSS' or instrume == 'PFIS')):
	    xsh = [0., xshift[0], 0., xshift[1]]
	    ysh = [0., yshift[0], 0., yshift[1]]
	    rot = [0., rotation[0], 0., rotation[1]]
	if (status == 0 and instrume == 'SALTICAM'):
	    xsh = [0., xshift[0], 0.]
	    ysh = [0., yshift[0], 0.]
	    rot = [0., rotation[0], 0]

# how many extensions?

	if (status == 0): nextend, status = saltkey.get('NEXTEND',struct[0],infile,logfile)

# CCD on-chip binning

	if (status == 0): 
	    xbin, ybin, status = saltkey.ccdbin(struct[0],infile,logfile)

# create temporary primary extension

	if (status == 0): 
	    outstruct = []
	    outstruct.append(struct[0])

# define temporary FITS file store tiled CCDs

	if (status == 0): tilefile, status = saltio.tmpfile(outpath,False,logfile)
	if (status == 0):
	    tilefile += 'tile.fits'
	    tilehdu = [None] * (nextend/2+1)
	    tilehdu[0] = pyfits.PrimaryHDU()
	    tilehdu[0].header = struct[0].header

# iterate over amplifiers, stich them to produce file of CCD images

	if (status == 0):
	    for i in range(nextend/2):
		hdu = i * 2 + 1
		amplifier = hdu%amplifiers
		if (amplifier == 0): amplifier = amplifiers

# read DATASEC keywords

		if (status == 0):
		    datasec1, status = saltkey.get('DATASEC',struct[hdu],  infile,logfile)
		    datasec2, status = saltkey.get('DATASEC',struct[hdu+1],infile,logfile)
		    xdsec1, ydsec1, status = saltstring.secsplit(datasec1,infile+'['+str(hdu)  +']',logfile)
		    xdsec2, ydsec2, status = saltstring.secsplit(datasec2,infile+'['+str(hdu+1)+']',logfile)

# read images

		if (status == 0):
		    imdata1, status = saltio.readimage(struct,hdu,  logfile)
		    imdata2, status = saltio.readimage(struct,hdu+1,logfile)

# tile 2n amplifiers to yield n CCD images

		if (status == 0):
		    outimage = numpy.empty((ydsec1[1],xdsec1[1]+xdsec2[1]),numpy.float32)
		    outimage.shape = ydsec1[1], xdsec1[1]+xdsec2[1]
		    outimage[ydsec1[0]-1:ydsec1[1],xdsec1[0]-1:xdsec1[1]] = \
			imdata1[ydsec1[0]-1:ydsec1[1],xdsec1[0]-1:xdsec1[1]]
		    outimage[ydsec2[0]-1:ydsec2[1],xdsec1[1]:xdsec1[1]+xdsec2[1]] = \
			imdata2[ydsec1[0]-1:ydsec1[1],xdsec1[0]-1:xdsec1[1]]

# size of new image

		    naxis1 = str(xdsec1[1] + xdsec2[1])
		    naxis2 = str(ydsec1[1])

# add image and keywords to HDU list

		    tilehdu[i+1] = pyfits.ImageHDU(outimage)
		    tilehdu[i+1].header = struct[hdu].header
		    tilehdu[i+1].header['DATASEC'] = '[1:' + naxis1 + ',1:' + naxis2 + ']'

# image tile log message #1

		if (status == 0):
		    message =  'SALTMOSAIC -- ' + os.path.basename(infile) + '[' + str(hdu) + ']['
		    message += str(xdsec1[0]) + ':' + str(xdsec1[1]) + ','
		    message += str(ydsec1[0]) + ':' + str(ydsec1[1]) + '] --> '
		    message += os.path.basename(tilefile) + '[' + str(i+1) + ']['
		    message += str(xdsec1[0]) + ':' + str(xdsec1[1]) + ','
		    message += str(ydsec1[0]) + ':' + str(ydsec1[1]) + ']'
		    saltprint.log(logfile,message,verbose)
		    message =  'SALTMOSAIC -- ' + os.path.basename(infile) + '[' + str(hdu+1) + ']['
		    message += str(xdsec1[0]) + ':' + str(xdsec1[1]) + ','
		    message += str(ydsec1[0]) + ':' + str(ydsec1[1]) + '] --> '
		    message += os.path.basename(tilefile) + '[' + str(i+1) + ']['
		    message += str(xdsec1[1]+1) + ':' + str(xdsec1[1]+xdsec2[1]) + ','
		    message += str(ydsec2[0]) + ':' + str(ydsec2[1]) + ']'
		    saltprint.log(logfile,message,verbose)

# write temporary file of tiled CCDs

	    if (status == 0): hdulist = pyfits.HDUList(tilehdu)
	    if (status == 0): hdulist.writeto(tilefile)

# iterate over CCDs, transform and rotate images

	yrot = [None] * 4
	xrot = [None] * 4
	if (status == 0):
	    tranfile = [' ']
	    tranhdu = [0]
	    for hdu in range(1,nextend/2+1):
		tranfile.append(' ')
		tranhdu.append(0)
		if (status == 0): tranfile[hdu], status = saltio.tmpfile(outpath,False,logfile)
		if (status == 0): tranfile[hdu] += 'tran.fits'
		ccd = hdu%nccds
		if (ccd == 0): ccd = nccds

# correct rotation for CCD binning

		yrot[ccd] = rot[ccd] * ybin / xbin
                xrot[ccd] = rot[ccd] * xbin / ybin
		dxshift = xbin * int(float(int(gap) / xbin) + 0.5) - gap

# transformation using geotran IRAF task

	        if (ccd != 2):
		    message  = '\nSALTMOSAIC -- geotran ' + tilefile + '[' + str(ccd) + '] ' + tranfile[hdu]
		    message += ' \"\" \"\" xshift=' + str((xsh[ccd]+(2-ccd)*dxshift)/xbin) + ' '
		    message += 'yshift=' + str(ysh[ccd]/ybin) + ' xrotation=' + str(xrot[ccd]) + ' '
		    message += 'yrotation=' + str(yrot[ccd]) + ' xmag=1 ymag=1 xmin=\'INDEF\''
		    message += 'xmax=\'INDEF\' ymin=\'INDEF\' ymax=\'INDEF\' ncols=\'INDEF\' '
		    message += 'nlines=\'INDEF\' verbose=\'no\' fluxconserve=\'yes\' nxblock=2048 '
		    message += 'nyblock=2048 interpolant=\'' + interp + '\' boundary=\'constant\' constant=0'
		    saltprint.log(logfile,message,verbose)
	            iraf.images.immatch.geotran(tilefile+"["+str(ccd)+"]",
						tranfile[hdu],
						"",
						"",
						xshift=(xsh[ccd]+(2-ccd)*dxshift)/xbin,
						yshift=ysh[ccd]/ybin,
						xrotation=xrot[ccd],
						yrotation=yrot[ccd],
						xmag=1,ymag=1,xmin='INDEF',xmax='INDEF',ymin='INDEF',
						ymax='INDEF',ncols='INDEF',nlines='INDEF',verbose='no',
						fluxconserve='yes',nxblock=2048,nyblock=2048,
						interpolant="linear",boundary="constant",constant=0)
		else:
		    tranfile[hdu] = tilefile
		    tranhdu[hdu] = hdu

# open outfile

	if (status == 0):
	    saltprint.log(logfile,' ',verbose)
	    outstruct = [None] * (nextend/nccds/2+1)
	    outstruct[0] = pyfits.PrimaryHDU()

# iterate over CCDs, stich them to produce a full image

	if (status == 0):
	    hdu = 0
	    for outhdu in range(1,nextend/nccds/2+1):
		for inhdu in range(1,nccds+1):
		    hdu += 1
		    instruct, status = saltio.openfits(tranfile[hdu],logfile)
		    inimage = instruct[tranhdu[hdu]].data
		    outstruct[0].header = struct[0].header

# read DATASEC keywords

		    if (status == 0): 
			datasec, status = saltkey.get('DATASEC',instruct[tranhdu[hdu]],
						      tranfile[hdu]+'['+str(tranhdu[hdu])+']',
						      logfile)
		    if (status == 0):
			xdsec, ydsec, status = saltstring.secsplit(datasec,
								   tranfile[hdu]+'['+str(tranhdu[hdu])+']',
								   logfile)

# define size and shape of final image

		    if (status == 0 and inhdu == 1):
			naxis1 = int(xdsec[1] * nccds + gap / xbin * (nccds - 1))
			naxis2 = int(ydsec[1])
			outimage = numpy.empty((naxis2,naxis1),numpy.float32)
			outimage.shape = naxis2, naxis1

# tile CCDs to yield mosaiced image

		    if (status == 0):
			x1 = int(xdsec[0] - 1 + (xdsec[1] + gap / xbin) * (inhdu - 1))
			x2 = int(xdsec[1] + (xdsec[1] + gap / xbin) * (inhdu - 1))
			y1 = int(ydsec[0] - 1)
			y2 = int(ydsec[1])
			outimage[y1:y2,x1:x2] = inimage[ydsec[0]-1:ydsec[1],xdsec[0]-1:xdsec[1]]

# image tile log message #1

		    if (status == 0):
			message =  'SALTMOSAIC -- ' + os.path.basename(tranfile[hdu])
			message += '[' + str(tranhdu[hdu]) + ']['
			message += str(xdsec[0]) + ':' + str(xdsec[1]) + ','
			message += str(ydsec[0]) + ':' + str(ydsec[1]) + '] --> '
			message += os.path.basename(outfile) + '[' + str(outhdu) + ']['
			message += str(x1+1) + ':' + str(x2) + ','
			message += str(y1+1) + ':' + str(y2) + ']'
			saltprint.log(logfile,message,verbose)

# add image and keywords to HDU list

		    if (status == 0 and inhdu == nccds):
			outstruct[outhdu] = pyfits.ImageHDU(outimage)
			outhead = outstruct[outhdu].header
			saltkey.copy(outstruct[outhdu],struct[0],'observat','South African Large Telescope')
			saltkey.copy(outstruct[outhdu],struct[0],'instrume','Detector instrument')
			saltkey.copy(outstruct[outhdu],struct[0],'detsw','Detector software version')
			saltkey.copy(outstruct[outhdu],struct[0],'detmode','Detector mode')
			saltkey.copy(outstruct[outhdu],struct[0],'object','Target object')
			saltkey.copy(outstruct[outhdu],struct[0],'ra','Target RA')
			saltkey.copy(outstruct[outhdu],struct[0],'dec','Target declination')
			saltkey.copy(outstruct[outhdu],struct[0],'objepoch','Epoch of object RA and Dec')
			saltkey.copy(outstruct[outhdu],struct[0],'equinox','Equinox of target coordinates')
			saltkey.copy(outstruct[outhdu],struct[0],'date-obs','Date of observation')
			saltkey.copy(outstruct[outhdu],struct[0],'utc-obs','UT start of observation')
			saltkey.copy(outstruct[outhdu],struct[0],'time-obs','UT start of observation')
			saltkey.copy(outstruct[outhdu],struct[0],'exptime','Exposure time (s)')
			saltkey.copy(outstruct[outhdu],struct[0],'detsize','Detector size')
			saltkey.copy(outstruct[outhdu],struct[0],'nccds','Number of CCDs in detector')
			saltkey.copy(outstruct[outhdu],struct[0],'ccdsum','On chip summation')
			saltkey.copy(outstruct[outhdu],struct[0],'gainset','CCD gain mode')
			saltkey.copy(outstruct[outhdu],struct[0],'rospeed','CCD readout speed')
			saltkey.copy(outstruct[outhdu],struct[0],'ccdtem','CCD temperture (K)')
			saltkey.copy(outstruct[outhdu],struct[0],'dewtem','Cold box temperature (C)')
			saltkey.copy(outstruct[outhdu],struct[0],'amptem','SDSU controller temperature (C)')
			saltkey.copy(outstruct[outhdu],struct[0],'centem','Cold end temperature (C)')
			saltkey.copy(outstruct[outhdu],struct[0],'dewpre','Dewar pressure')
			saltkey.copy(outstruct[outhdu],struct[0],'pcs-ver','Version of PFIS control software')
			saltkey.copy(outstruct[outhdu],struct[0],'proposer','Proposal Principle Investigator')
			saltkey.copy(outstruct[outhdu],struct[0],'propid','Proposal identification number')
			saltkey.copy(outstruct[outhdu],struct[0],'posang','Planned target position angle')
			saltkey.copy(outstruct[outhdu],struct[0],'ccdtype','CCD type')
			saltkey.copy(outstruct[outhdu],struct[0],'lampid','Name of lamp')
			saltkey.copy(outstruct[outhdu],struct[0],'calfilt','Name of filter')
			saltkey.copy(outstruct[outhdu],struct[0],'calnd','Neutral density setting percentage')
			saltkey.copy(outstruct[outhdu],struct[0],'trkx','Tracker x (m)')
			saltkey.copy(outstruct[outhdu],struct[0],'trky','Tracker y (m)')
			saltkey.copy(outstruct[outhdu],struct[0],'trkz','Tracker z (m)')
			saltkey.copy(outstruct[outhdu],struct[0],'trkphi','Tracker Phi (deg)')
			saltkey.copy(outstruct[outhdu],struct[0],'trktheta','Tracker Theta (deg)')
			saltkey.copy(outstruct[outhdu],struct[0],'trkrho','Tracker Rho (deg)')
			saltkey.copy(outstruct[outhdu],struct[0],'intfr','Interferometer reading')
			saltkey.copy(outstruct[outhdu],struct[0],'colphi','Collimator Phi (deg)')
			saltkey.copy(outstruct[outhdu],struct[0],'coltheta','Collimator Theta (deg)')
			saltkey.copy(outstruct[outhdu],struct[0],'telha','HA of telescope')
			saltkey.copy(outstruct[outhdu],struct[0],'telra','Telescope RA')
			saltkey.copy(outstruct[outhdu],struct[0],'teldec','Telescope declination')
			saltkey.copy(outstruct[outhdu],struct[0],'telpa','Telescope position angle')
			saltkey.copy(outstruct[outhdu],struct[0],'telrho','Position angle of Rho stage')
			saltkey.copy(outstruct[outhdu],struct[0],'telaz','Telescope azimuth')
			saltkey.copy(outstruct[outhdu],struct[0],'telalt','Telescope elevation')
			saltkey.copy(outstruct[outhdu],struct[0],'teltemp','Telescopr temperature (C)')
			saltkey.copy(outstruct[outhdu],struct[0],'payltemp','Payload temperature (C)')
			saltkey.copy(outstruct[outhdu],struct[0],'pellicle','Position of SALTICAM pellicle')
			saltkey.copy(outstruct[outhdu],struct[0],'instport','Instrument port')
			saltkey.copy(outstruct[outhdu],struct[0],'obsmode','Config state machine state')
			saltkey.copy(outstruct[outhdu],struct[0],'maskid','Slit mask bar code')
			saltkey.copy(outstruct[outhdu],struct[0],'focus','Focus position (steps)')
			saltkey.copy(outstruct[outhdu],struct[0],'grating','Commanded grating station')
			saltkey.copy(outstruct[outhdu],struct[0],'gr-angle','Grating rotation angle')
			saltkey.copy(outstruct[outhdu],struct[0],'filter','Filter bar code')
			saltkey.copy(outstruct[outhdu],struct[0],'ar-angle','Articulation angle')
			saltkey.copy(outstruct[outhdu],struct[0],'sal-tlm','File last updated by SALT tools')
			saltkey.copy(outstruct[outhdu],struct[0],'sal-edt','Keywords updated by SALTEDTKY')
			saltkey.copy(outstruct[outhdu],struct[0],'pprepare','File prepared for IRAF')
			saltkey.copy(outstruct[outhdu],struct[0],'sprepare','File prepared for IRAF')
			saltkey.copy(outstruct[outhdu],struct[0],'pgain','Image is corrected for amplifier gain')
			saltkey.copy(outstruct[outhdu],struct[0],'pxtalk','Amplifiers corrected for crosstalk')
			saltkey.copy(outstruct[outhdu],struct[0],'pbias','Bias subtracted image')
			outhead.update(keymosaic,time.asctime(time.localtime()),'Image is a mosaic')
			#outhead.add_history(history)
			
# housekeeping keywords

			if (status == 0): status = saltkey.put('SAL-TLM',time.asctime(time.localtime()),
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('EXTNAME','SCI','Extension name',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('EXTVER',outhdu,'Extension number',
							       outstruct[outhdu],outfile,logfile)

# WCS keywords

			if (status == 0): status = saltkey.new('CRPIX1',1,'WCS: X reference pixel',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CRPIX2',1,'WCS: Y reference pixel',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CRVAL1',float(xbin),
							       'WCS: X reference coordinate value',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CRVAL2',float(ybin),
							       'WCS: Y reference coordinate value',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CDELT1',float(xbin),'WCS: X pixel size',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CDELT2',float(ybin),'WCS: Y pixel size',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CTYPE1','pixel','X type',
							       outstruct[outhdu],outfile,logfile)
			if (status == 0): status = saltkey.new('CTYPE2','pixel','Y type',
							       outstruct[outhdu],outfile,logfile)
			    
# close temporary FITS file

		    if (status == 0): status = saltio.closefits(instruct,logfile)

# close input FITS file

	if (status == 0): status = saltio.closefits(struct,logfile)

# housekeeping keywords

	if (status == 0): status = saltkey.put('SAL-TLM',time.asctime(time.localtime()),
					       outstruct[0],outfile,logfile)
	if (status == 0): status = saltkey.new(keymosaic,time.asctime(time.localtime()),
					       'Amplifiers have been mosaiced',outstruct[0],outfile,logfile)
	if (status == 0): 
	        history = 'SALTMOSAIC -- '
                history += 'images='+infile+' '
                history += 'outimages='+outfile+' '
		status = saltkey.history(outstruct[0],history,outfile,logfile)

# clobber old files
	    
	if (status == 0 and clobber and os.path.isfile(outfile)): 
	    status = saltio.delete(outfile,False,logfile)

# cleanup temporary files

	if (status == 0 and cleanup):
	    for file in tranfile:
		if os.path.isfile(file):
		    status = saltio.delete(file,False,logfile)
	    if os.path.isfile(file):
		status = saltio.delete(tilefile,False,logfile)

# write FITS file of mosaiced image

	if (status == 0): outstruct = pyfits.HDUList(outstruct)
	if (status == 0): outstruct.writeto(outfile)
	if (status == 0): 
	    message = '\nSALTMOSAIC -- created ' + os.path.basename(outfile) + ' at'
	    saltprint.time(message,logfile,verbose)
	    saltprint.log(logfile,' ',verbose)
	ifil += 1

# end time

    if (status == 0):
	saltprint.time('SALTMOSAIC -- completed at',logfile,verbose)
    else:
	saltprint.time('\nSALTMOSAIC -- aborted at',logfile,verbose)

# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("saltred$saltmosaic.par") 
t = iraf.IrafTaskFactory(taskname="saltmosaic",value=parfile,function=saltmosaic, pkgname='saltred')
