#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# Martin Still (SAAO)    1.0          21 Jul 2006

# saltprepare...

from pyraf import iraf 
from pyraf.iraf import pysalt
import saltprint, saltio, saltkey
import os, string, sys, glob, pyfits, time

# -----------------------------------------------------------
# core routine

def saltprepare(images,rawpath,outimages,outpref,clobber,logfile,verbose,status):

    infiles = []
    outfiles = []
    status = 0

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    history = 'SALTPREPARE -- '
    history += 'images='+images+' '
    history += 'outimages='+outimages+' '
    history1 = 'rawpath='+rawpath+' '
    history1 += 'outpref='+outpref+' '
    yn = 'n'
    if (clobber): yn = 'y'
    history1 += 'clobber='+yn+' '
    history1 += 'logfile='+logfile+' '
    yn = 'n'
    if (verbose): yn = 'y'
    history1 += 'verbose='+yn
    history += history1
    saltprint.log(logfile,history,verbose)

# start time

    saltprint.time('\nSALTPREPARE -- started at',logfile,verbose)
    saltprint.log(logfile,' ',verbose)

# check the rawpath exists and ends with a "/"

    rawpath, status = saltio.pathexists(rawpath,logfile)

# is the input file specified?
   
    if (status == 0): status = saltio.filedefined('Input',images,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
	status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
	infiles, status = saltio.listparse('Raw image',images,'',infiles,rawpath,logfile)

# check raw input files exist

    if (status == 0): status = saltio.filesexist(infiles,rawpath,'r',logfile)

# load output name list: @list, * and comma separated
   
    outimages = outimages.strip()
    outpref = outpref.strip()
    if (status == 0 and len(outpref) == 0 and len(outimages) == 0):
	message = 'ERROR -- SALTPREPARE: output file(s) not specified'
	status = saltprint.err(logfile,message)

# test output @filelist exists

    if (status == 0 and len(outimages) > 0 and outimages[0] == '@'): 
	status = saltio.listexists('Output',outimages,logfile)

# parse list of output files

    if (status == 0):
	outfiles, status = saltio.listparse('Output image',outimages,outpref,infiles,'',logfile)

# are input and output lists the same length?

    if (status == 0): status = saltio.comparelists(infiles,outfiles,'Input','output',logfile)

# do the output files already exist?

    if (status == 0 and not clobber): status = saltio.filesexist(outfiles,'','w',logfile)

# open each raw image file

    ifil = 0
    for imageitem in infiles:
	infile = rawpath+imageitem
	outfile = outfiles[ifil]
	if (status == 0): struct, status = saltio.openfits(infile,logfile)

# prepare file

	if (status == 0): 
             history = 'SALTPREPARE -- '
             history += 'images='+infile+' '
             history += 'outimages='+outfile+' '
	     history += history1
	     struct, status = prepare(struct,infile,history,logfile,status)

# if clobbering, delete old output file

	if (status == 0 and os.path.isfile(outfile)): saltio.delete(outfile,'no',logfile)
	    
# write FITS file

	if (status == 0): status = saltio.writefits(struct,outfile,logfile)
	if (status == 0): status = saltio.closefits(struct,logfile)
	message = 'SALTPREPARE -- ' + infile+' => '+outfile
	if (status == 0): saltprint.log(logfile,message,verbose)
	ifil += 1

# end time

    if (status == 0):
	saltprint.time('\nSALTPREPARE -- completed at',logfile,verbose)
    else:
	saltprint.time('\nSALTPREPARE -- aborted at',logfile,verbose)

# -----------------------------------------------------------
# prepare FITS file for processing

def prepare(struct,infile,history,logfile,status):

# identify instrument

    if (status == 0): instrume,keyprep,keygain,keybias,keyxtalk,keyslot,status = \
	saltkey.instrumid(struct,infile,logfile)

# has file been prepared already?

    if (status == 0): status = 2
    if (status == 2):
	try:
	    key = struct[0].header[keyprep]
	except:
	    status = 0
    if (status == 2):
	message = 'ERROR -- SALTPREPARE: File ' + infile 
	message += ' has already been prepared'
	status = saltprint.log(logfile,message)

# number of image HDU

    nextend = 0
    while (status == 0):
	try:
	    if (struct[nextend+1].header['XTENSION'] == 'IMAGE'): nextend += 1
	except:
	    break
    nccds,status = saltkey.get('NCCDS',struct[0],infile,logfile)
    if (nextend%(nccds*2) != 0):
	message = 'ERROR -- SALTPREPARE: Number of file extensions and '
	message += 'number of amplifiers are not consistent in ' + infile
	status = saltprint.err(logfile,message)
    if (status == 0): status = saltkey.new('NSCIEXT',nextend,'Number of science extensions',
			 struct[0],infile,logfile)
    if (status == 0): status = saltkey.new('NEXTEND',nextend,'Number of data extensions',
			 struct[0],infile,logfile)

# housekeeping keywords

    if (status == 0): status = saltkey.housekeeping(struct[0],keyprep, 'File prepared for IRAF', history,infile,logfile)

# keywords for image extensions

    if (status == 0): 
	for i in range(nextend):
	    hdu = i + 1
	    status = saltkey.new('EXTNAME','SCI','Extension name',struct[hdu],infile,logfile)
	    status = saltkey.new('EXTVER',hdu,'Extension number',struct[hdu],infile,logfile)

    return struct, status

# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("saltred$saltprepare.par") 
t = iraf.IrafTaskFactory(taskname="saltprepare",value=parfile,function=saltprepare, pkgname='saltred')
