#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# Martin Still (SAAO)    0.2          13 Nov 2006
# S M Crawford (SAA0)    0.3          11 Feb 2008
#
# SLOTPHOT is a tool to do photometry on a serious of data 
# obtained with slotmode configuration of salticam.  
# After reading in the arguments, the algorithm is the
# following:
# 1.  Open th fits file
# 2.  Subtact a smooth background  
# 3.  Centroid the comparison star to acount for drift
# 4.  Photometry on Comparison and target star
# 5.  Write data
# 6.  Plot data [optional]
# 7.  Write fits [optional]
#
# One of the basic assumptions of this program are that 
# the target and comparison star are on the same amplifier
#

from pyraf import iraf 
import saltprint, saltkey, saltio, salttime
import slottool

import numpy
from pylab import *

# -----------------------------------------------------------
# core routine

def slotphot(images,outfile,srcfile, newfits, phottype, subbacktype, sigback, mbin, sorder,  iter,  \
	     sigdet, contpix, vignfile,ignorexp, reltime,plotdata,plotfreq, clobber,logfile,verbose,status):

    import os, time, saltio, re, pyfits


#set up the variables
    status = 0
    entries = []
    vig_lo = {}
    vig_hi = {}
    amp = {}
    x = {}
    y = {}
    x_o = {}
    y_o = {}
    r = {}
    br1 = {}
    br2 = {}
    hour = 0
    min = 0
    sec = 0.
    time0 = 0.
    nframes = 0
    bin=mbin
    order=sorder
    


# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    message = 'SLOTPHOT -- '
    message += 'images='+images+' '
    message += 'outfile='+outfile+' '
    message += 'srcfile='+srcfile+' '
    message += 'newfits='+newfits+' '
    message += 'subbacktype='+subbacktype+' '
    message += 'phottype='+phottype+' '
    message += 'sigback=%f ' % sigback
    message += 'bin=%i ' % bin
    message += 'iter=%i '% iter
    message += 'order=%i ' % order
    message += 'sigdet=%f ' % sigdet
    message += 'contpix=%i ' % contpix
    message += 'ignorexp=%i ' % ignorexp 
    message += 'vignfile='+vignfile+' '
    yn = 'n'
    if (reltime): yn = 'y'
    message += 'reltime='+yn+' '
    yn = 'n'
    if (plotdata): yn = 'y'
    message += 'plot='+yn+' '
    yn = 'n'
    message += 'plotfreq='+str(plotfreq)+' '
    if (clobber): yn = 'y'
    message += 'clobber='+yn+' '
    message += 'logfile='+logfile+' '
    yn = 'n'
    if (verbose): yn = 'y'
    message += 'verbose='+yn+'\n'
    saltprint.log(logfile,message,verbose)

# start time

    saltprint.time('SLOTPHOT started at  ',logfile,verbose)

# is the input file specified?
   
    if (status == 0): status = saltio.filedefined('Input',images,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# is the output file specified?
   
    if (status == 0): status = saltio.filedefined('Output',outfile,logfile)

# check output file does not exist, optionally remove it if it does exist

    if (status == 0): 
	status = saltio.filedoesnotexist(outfile,'no',logfile)
        if (status == 1 and clobber): 
	  os.remove(outfile)
	  status = 0
        elif (status == 1 and not clobber):
	  message = 'ERROR -- SLOTPHOT: File ' + outfile + ' already exists, use clobber=y'
	  status = saltprint.err(logfile,message)

# open output ascii file

    if (status == 0):
	try:
	    lc = open(outfile,'a')
	except:
	    message = 'ERROR -- SLOTPHOT: Cannot open ouput file ' + outfile
	    status = saltprint.err(message,logfile)

# is the extraction region defintion file specified?
   
    if (status == 0): status = saltio.filedefined('Extraction region defintion',srcfile,logfile)

# check extraction region defintion file exists

    srcfile = srcfile.strip()
    if (status == 0): status = saltio.fileexists(srcfile,logfile)

# read extraction region defintion file

    if (status == 0):
	region, status = saltio.openascii(srcfile,'r',logfile)
	for line in region:
	    if (len(line.strip()) > 0 and line[0] != '#'):
		line = line.rstrip('\r\n')
		line = line.lstrip()
		line = re.sub("\s+",",",line)
		line.rstrip(',')
		entries = line.split(',')
		if (entries[0] == '1' or entries[0]=='t'): entries[0] = 'target'
		if (entries[0] == '2' or entries[0]=='c'): entries[0] = 'comparison'
		if (entries[0] == 'target' or entries[0] == 'comparison'):
		    amp[entries[0]] = int(entries[1])
		    x[entries[0]] = int(float(entries[2]))
		    y[entries[0]] = int(float(entries[3]))
		    x_o[entries[0]] = int(float(entries[2]))
		    y_o[entries[0]] = int(float(entries[3]))
		    r[entries[0]] = int(float(entries[4]))
		    br1[entries[0]] = int(float(entries[5]))
		    br2[entries[0]] = int(float(entries[6]))
	status = saltio.closeascii(region,logfile)

# is extraction region data complete

    if status==0:
     if (len(amp) != 2 or len(x) != 2 or len(y) != 2 or len(r) != 2 or len(br1) != 2 or len(br2) != 2):
	message = 'ERROR: SLOTPHOT -- extraction region definition file '+srcfile
	message += ' does not have an acceptable format.'
	status = saltprint.err(logfile,message)

# read vignetting region defintion file

    if (status == 0):
	region, status = saltio.openascii(vignfile,'r',logfile)
	for line in region:
	    if (len(line.strip()) > 0 and line[0] != '#'):
		line = line.rstrip('\r\n')
		line = line.lstrip()
		line = re.sub("\s+",",",line)
		line.rstrip(',')
		entries = line.split(',')
		vig_lo[entries[0]] = entries[1]
		vig_hi[entries[0]] = entries[2]
	status = saltio.closeascii(region,logfile)

# set the writenewfits parameter

    if status==0:
	if not newfits or newfits=='none':
	    writenewfits=False
	else:
	    writenewfits=newfits


# get time of first exposure and basic information about the observations


    if (status == 0):

	infile=infiles[0]
	struct, status = saltio.openfits(infile,logfile)

        # identify instrument
        if (status ==0 ):
	   instrume,keyprep,keygain,keybias,keyxtalk,keyslot,status = saltkey.instrumid(struct,infile,logfile)

	# how many extensions?
    	if (status == 0): 
		nextend,status = saltkey.get('NEXTEND',struct[0],infile,logfile)
		if nextend < amp['comparison']:
		   message='SALTPHOT--ERROR:  Insufficient number of extensions in %s' % (infile)
		   status=saltprint.err(logfile, message)

	# how many amplifiers?
    	if (status == 0): 
	   amplifiers,status = saltkey.get('NCCDS',struct[0],infile,logfile)
	   amplifiers = int(2*float(amplifiers))
	   nframes = nextend/amplifiers

	# image size
        if (status == 0):
	   naxis1,status = saltkey.get('NAXIS1',struct[amp['comparison']],infile,logfile) 
	   naxis2,status = saltkey.get('NAXIS2',struct[amp['comparison']],infile,logfile) 

	# CCD binning
        if (status == 0):
	  ccdsum,status = saltkey.get('CCDSUM',struct[0],infile,logfile) 
	  binx = int(ccdsum.split(' ')[0])
	  biny = int(ccdsum.split(' ')[1])


	#Identify the time of the observations
	hdu = 1
	try:
	  time0=slottool.getobstime(struct[hdu], infile+'['+str(hdu)+']', logfile)
	  dateobs, status = saltkey.get('DATE-OBS',struct[hdu],infile,logfile)
	  dateobs = dateobs.replace('-','/')
	except:
	   message='ERROR--SLOTPHOT:  No time or obsdate in first image'
	   status=saltprint.err(logfile,message)

        # If a total file is to written out, create it and update it
        if status==0 and writenewfits:
	   #delete the file if clobber is on and the file exists
	   if (status == 0 and os.path.isfile(writenewfits)):
	     if clobber:
              status = saltio.delete(writenewfits,'no',logfile)
	     else:
	      message='ERROR -- SLOTPHOT: Output fits file ' + writenewfits + 'already exists. '
	      message += 'Use Clobber=yes to overwrite file'
              status = saltprint.err(logfile,message)
	
	   #create the output file 
	   if status==0:
	    try:
              hdu = pyfits.PrimaryHDU()
              hdu.header=struct[0].header
              hduList = pyfits.HDUList(hdu)
              hduList.verify()
              hduList.writeto(writenewfits)
            except:
              message  = 'ERROR -- SLOTPHOT: Could not create new fits file named ' + writenewfits
              status = saltprint.err(logfile,message)
	
	   #read it back in for updating
	   if status==0:
            try:
                  hduList = pyfits.open(writenewfits,mode='update')
            except:
                  message = 'ERROR -- SLOTPHOT:  Canot open  ' + outfile
                  status = saltprint.err(logfile,message)

# set up the arrays
    if status==0:
	ntotal=nframes*len(infiles)
	time=numpy.zeros(ntotal ,dtype='float')-1.0
	dx=numpy.zeros(ntotal ,dtype='float')
	dy=numpy.zeros(ntotal ,dtype='float')
	tflux=numpy.zeros(ntotal ,dtype='float')
	terr =numpy.zeros(ntotal ,dtype='float')
	cflux=numpy.zeros(ntotal ,dtype='float')
	cerr =numpy.zeros(ntotal ,dtype='float')
	ratio=numpy.zeros(ntotal ,dtype='float')
	rerr =numpy.zeros(ntotal ,dtype='float')
	

#start displaying the amount complete

    if status==0 and plotdata:
	#compfig  = figure(figsize=(3,1.0),dpi=72)
	compdisp = axes([0.05,0.15,0.2,0.2])
	title('Percentage Complete')
	plot(tflux,terr)
	compdisp .set_xlim(0,1)
        compdisp.set_ylim(0,1)
	setp(gca(),xticks=[],yticks=[])
	ion()
	ylabel('nope')
	ioff()




# open each raw image file

    j=0
    ftime=0
    if status==0:
     for infile in infiles:

	if (status == 0 ): struct, status = saltio.openfits(infile,logfile)

        # has file been cleaned?
        if (status == 0): cstatus=saltkey.clean(struct,infile,keyslot,keygain,logfile)

	#Skip through the frames and process each frame individually
	for i in range(nframes):
          if not (infile==infiles[0] and i < ignorexp):
	   pstatus=0
	   hdu=amp['comparison']+i*amplifiers
	   try:
	     header=struct[hdu].header
	     array=struct[hdu].data
	     array=array*1.0
	   except:
	     message='SLOTPHOT--ERROR:  Unable to open extension %i in image %s' % (hdu, infile)
	     pstatus=saltprint.err(logfile,message)


	   #start the analysis of each frame
	   if pstatus==0:
             # get the time
	     time[j]=slottool.getobstime(struct[hdu],infile+'['+str(hdu)+']',logfile)

             # gain and readout noise
	     try:
	       gain=float(header['GAIN'])
	     except:
	       gain=1
	       message='WARNING--SLOTPHOT:  Gain not specified in image header'
	       saltprint.log(logfile,message,verbose)
	     try:
	       rdnoise=float(header['RDNOISE'])
	     except:
	       rdnoise=0
	       message='WARNING--SLOTPHOT:  RDNOISE not specified in image header'
	       saltprint.log(logfile,message,verbose)

	   # background subtraction
	   if pstatus==0 and not subbacktype=='none':
	     array,status = slottool.subbackground(array, sigback, bin, order, iter, subbacktype, logfile)

	   #x-y fit to the comparison star and update the x,y values
	   if pstatus==0:
	     carray, fx,fy,dstatus = slottool.finddrift(array, x['comparison'], y['comparison'], r['comparison'], \
					naxis1, naxis2, sigdet, contpix, sigback, iter, logfile)
	     if fx > -1  and fy > -1:
              if fx < naxis1 and fy < naxis2:
	       dx[j]=x['comparison']-fx
	       dy[j]=y['comparison']-fy
	       x['comparison']=fx
	       y['comparison']=fy
	       x['target']=x['target']-dx[j]
	       y['target']=y['target']-dy[j]
	      else:
	       dx[j]=0
	       dy[j]=0
	       x['comparison']=x_o['comparison']
               y['comparison']=y_o['comparison']
               x['target']=x_o['target']
               y['target']=y_o['target']
	     else:
	      message  = 'WARNING--SLOTPHOT:  No comparison object found in image file ' + infile
	      message += ' on extension %i \n' % hdu
	      dstatus=saltprint.err(logfile, message)
	
	   #do photometry
	   if pstatus==0:
		
	     tflux[j],terr[j],cflux[j],cerr[j],ratio[j],rerr[j],status =  \
		slottool.dophot(phottype, array, x, y, r, br1, br2, gain, rdnoise, naxis1, naxis2, verbose, logfile)

	   #record results
	   if pstatus==0:
             status=slottool.writedataout(lc, j+1, time[j], x, y, tflux[j], terr[j], cflux[j],cerr[j],ratio[j],rerr[j], \
		time0, reltime, logfile, verbose)
	
	   #display the completeness of the processing 
	   if pstatus==0 and plotdata:
	      pass
	
	   #write the fits file
	   if pstatus==0 and writenewfits: 
		hdue = pyfits.ImageHDU(array)
		hdue.header=header
		hduList.append(hdue)

	   j +=1
	   if status==1: print j

        # close FITS file		
	status = saltio.closefits(struct,logfile)

# close the output fits file:
	
    if status==0 and writenewfits:
	  try:
           hduList.flush()
           hduList.close()
          except:
           message = 'ERROR: SLOTPHOT-- Failed to write ' + writenewfits
           status = saltprint.err(logfile,message)


# close output ascii file

    if (status == 0):
	try:
	    lc.close()
	except:
	    message = 'ERROR -- SLOTPHOT: Cannot close ouput file ' + outfile
	    status = saltprint.err(message,logfile)

# end time

    if (status == 0):
	saltprint.time('SLOTPHOT completed at',logfile,verbose)
    else:
	saltprint.time('SLOTPHOT aborted at',logfile,verbose)


# ----------------------------------------------------------------------------
# Display of the completeness  

class CompDisplay():

    def __init__(ntot):
	self.ntot=ntot
	self.root=Tk.Tk()


# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("slottools$slotphot.par") 
t = iraf.IrafTaskFactory(taskname="slotphot",value=parfile,function=slotphot,pkgname='slottools')
